"""Functions for system-related tasks (currently just registering/deregistering the executor manifest)."""

import json
import logging
import os
import sys
import typing

from .interpreter import Interpreter

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

MANIFEST_FILE_NAME = 'pyla.json'
EXECUTORS_DIR_NAME = 'executors'


class ManifestManager:
    """Register or deregister a manifest file."""

    @staticmethod
    def get_home_dir() -> typing.Optional[str]:
        """Get the "home" directory of the current user, (or `APPDATA` dir on windows)."""
        os_name = sys.platform.lower()

        if os_name == 'windows':
            return os.getenv('APPDATA')

        return os.getenv('HOME')

    def user_dir(self) -> str:
        """
        Get the current user's Stencila data directory.

        This is the directory that Stencila configuration settings, such as the installed Stencila hosts, and document
        buffers get stored.
        """
        os_name = sys.platform.lower()

        home_dir = self.get_home_dir()

        if not home_dir:
            raise RuntimeError('Could not determine home directory from environment.')

        if os_name == 'darwin':
            return os.path.join(home_dir, 'Library', 'Application Support', 'Stencila')

        if os_name == 'linux':
            return os.path.join(home_dir, '.stencila')

        if os_name == 'windows':
            return os.path.join(home_dir, 'Stencila')

        return os.path.join(home_dir, 'stencila')

    def manifest_dir(self) -> str:
        """Get the directory in which execution registration manifests are stored."""
        return os.path.join(self.user_dir(), EXECUTORS_DIR_NAME)

    def manifest_path(self) -> str:
        """Get the path the the execution registration manifest for this type of executor (Python)."""
        return os.path.join(self.manifest_dir(), MANIFEST_FILE_NAME)

    def register(self) -> None:
        """
        Write the registration manifest (`MANIFEST`) to the manifest path.

        The command path (i.e. path to python) is set to the currently executing python binary, before right. This is to
        provide compatibility with virtual environments. The means that if you re-run the register command with
        different python binaries a different manifest will be written out.
        """
        os.makedirs(self.manifest_dir(), exist_ok=True)
        with open(self.manifest_path(), 'w') as manifest_file:
            json.dump(Interpreter.MANIFEST, manifest_file, indent=True)
        LOGGER.info('Manifest saved to \'%s\'', self.manifest_path())

    def deregister(self) -> None:
        """
        Remove the manifest file created with `register`.

        Does not fail if the manifest file does not exist.
        """
        if os.path.exists(self.manifest_path()):
            os.unlink(self.manifest_path())
            LOGGER.info('Deleted manifest at path \'%s\'', self.manifest_path())
        else:
            LOGGER.warning('Not deregistering as file \'%s\' does not exist', self.manifest_path())


def register() -> None:
    """Register the `MANIFEST` as defined by `EXECUTORS_DIR_NAME` and `MANIFEST_FILE_NAME`."""
    ManifestManager().register()


def deregister() -> None:
    """Deregister the `MANIFEST` as defined by `EXECUTORS_DIR_NAME` and `MANIFEST_FILE_NAME`."""
    ManifestManager().deregister()



"""
Handles parsing of CodeChunks to extract their properties and information about variables and functions used/defined.
"""
import ast
import logging
import traceback
import typing

from stencila.schema.types import SoftwareSourceCode, Function, Variable, CodeError, CodeChunk, SchemaTypes, \
    BooleanSchema, StringSchema, IntegerSchema, NumberSchema, ArraySchema, TupleSchema, CodeExpression, Parameter

ImportsType = typing.List[typing.Union[str, SoftwareSourceCode]]
OptionalStringList = typing.Optional[typing.List[str]]

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# pylint: disable=R0902
class CodeChunkParseResult:
    """The result of parsing a `CodeChunk`."""
    chunk_ast: typing.Optional[ast.Module]
    _imports: ImportsType
    _assigns: typing.List[str]
    _declares: typing.List[typing.Union[Function, Variable]]
    _alters: typing.List[str]
    _uses: typing.List[str]
    _reads: typing.List[str]
    error: typing.Optional[CodeError]

    # pylint: disable=R0913
    def __init__(self,
                 chunk_ast: typing.Optional[ast.Module] = None,
                 imports: typing.Optional[ImportsType] = None,
                 assigns: OptionalStringList = None,
                 declares: OptionalStringList = None,
                 alters: OptionalStringList = None,
                 uses: OptionalStringList = None,
                 reads: OptionalStringList = None,
                 error: typing.Optional[CodeError] = None):
        self.chunk_ast = chunk_ast
        self.imports = imports or []
        self.assigns = assigns or []
        self.declares = declares or []
        self.alters = alters or []
        self.uses = uses or []
        self.reads = reads or []
        self.error = error

    def combined_code_imports(self, existing_imports: typing.Optional[ImportsType]) -> typing.Optional[ImportsType]:
        """
        Combine a list of existing imports (from a CodeChunk) with the parsed `imports`.

        `self.imports` will be combined with `existing_imports`, with duplicates removed, unless `existing_imports` has
        an empty string semaphore which indicates no new imports should be added.
        """
        if existing_imports is None or self.imports is None:
            return self.imports

        if '' in existing_imports:
            return existing_imports

        imports = list(existing_imports)  # copy the list

        for imp in self.imports:
            if imp not in imports:
                imports.append(imp)

        return imports

    @property
    def imports(self) -> typing.Optional[ImportsType]:
        """Get the `imports` list or `None` if the list is empty."""
        if not self._imports:
            return None

        return self._imports

    @imports.setter
    def imports(self, imports: ImportsType) -> None:
        self._imports = imports

    @property
    def assigns(self) -> OptionalStringList:
        """Get the `assigns` list or `None` if the list is empty."""
        if not self._assigns:
            return None
        return self._assigns

    @assigns.setter
    def assigns(self, assigns: typing.List[str]) -> None:
        self._assigns = assigns

    @property
    def declares(self) -> typing.Optional[typing.List[typing.Union[Function, Variable]]]:
        """Get the `declares` list or `None` if the list is empty."""
        if not self._declares:
            return None
        return self._declares

    @declares.setter
    def declares(self, declares: typing.List[typing.Union[Function, Variable]]) -> None:
        self._declares = declares

    @property
    def alters(self) -> OptionalStringList:
        """Get the `alters` list or `None` if the list is empty."""
        if not self._alters:
            return None
        return self._alters

    @alters.setter
    def alters(self, alters: typing.List[str]) -> None:
        self._alters = alters

    @property
    def uses(self) -> OptionalStringList:
        """Get the `uses` list or `None` if the list is empty."""
        if not self._uses:
            return None
        return self._uses

    @uses.setter
    def uses(self, uses: typing.List[str]) -> None:
        self._uses = uses

    @property
    def reads(self) -> OptionalStringList:
        """Get the `reads` list or `None` if the list is empty."""
        if not self._reads:
            return None
        return self._reads

    @reads.setter
    def reads(self, reads: typing.List[str]) -> None:
        self._reads = reads

    def to_dict(self):
        """Return all attributes as a dictionary.."""
        return {
            'chunk_ast': self.chunk_ast,
            'imports': self.imports,
            'assigns': self.assigns,
            'declares': self.declares,
            'alters': self.alters,
            'uses': self.uses,
            'reads': self.reads,
            'error': self.error
        }


class CodeChunkExecution(typing.NamedTuple):
    """
    Combination of a `CodeChunk` and its parse result.

    This is so the AST does not have to be parsed twice (once during parsing and again during execution).
    """
    code_chunk: CodeChunk
    parse_result: CodeChunkParseResult


def annotation_name_to_schema(name: typing.Optional[str]) -> typing.Optional[SchemaTypes]:
    """Parse a Python annotation string (basically a type name) and convert to a `Schema` type."""
    if name is None:
        return None

    return {
        'bool': BooleanSchema(),
        'str': StringSchema(),
        'int': IntegerSchema(),
        'float': NumberSchema(),
        'list': ArraySchema(),
        'tuple': TupleSchema()
    }.get(name)


def mode_is_read(mode: str) -> bool:
    """
    Determine if an open mode is a read. Opening a file with `w+` or `a+` allows read and write.
    """
    return 'r' in mode or '+' in mode


def parse_open_filename(open_call: ast.Call) -> typing.Optional[str]:
    """
    Return the filename used in an `open` call (whether it's define positionally or with kwarg).

    If the filename is a variable, or the mode is not a read (or is a variable and thus can't be determined) then return
     `None`.
    """
    filename = None

    if hasattr(open_call, 'args'):
        if len(open_call.args) >= 1:
            if not isinstance(open_call.args[0], ast.Str):
                return None
            filename = open_call.args[0].s

            if len(open_call.args) >= 2:
                if not isinstance(open_call.args[1], ast.Str):
                    return None

                if not mode_is_read(open_call.args[1].s):
                    return None

    if hasattr(open_call, 'keywords'):
        for keyword in open_call.keywords:
            if not isinstance(keyword.value, ast.Str):
                continue

            if keyword.arg == 'file':
                filename = keyword.value.s

            if keyword.arg == 'mode':
                if not mode_is_read(keyword.value.s):
                    return None

    return filename


def exception_to_code_error(exception: Exception) -> CodeError:
    """Convert an `Exception` to a `CodeError` entity."""
    return CodeError(type(exception).__name__, message=str(exception), trace=traceback.format_exc())


def set_code_error(code: typing.Union[CodeChunk, CodeExpression],
                   error: typing.Union[Exception, CodeError]) -> None:
    """
    Add the `CodeError` to a `CodeChunk` or `CodeExpression` `errors` list.

    If an `Exception` is passed then it is converted to a `CodeError`.
    """
    if code.errors is None:
        code.errors = []

    if isinstance(error, Exception):
        error = exception_to_code_error(error)

    code.errors.append(error)


def simple_code_chunk_parse(code: CodeChunk) -> CodeChunkExecution:
    """
    "Build a CodeChunkExecution from CodeChunk.

    This is the most basic information that is needed to execute a CodeChunk in the interpreter.
    """
    parser = CodeChunkParser()
    cc_result = parser.parse(code)

    if cc_result.error:
        set_code_error(code, cc_result.error)

    return CodeChunkExecution(code, cc_result)


class CodeChunkParser:
    """Parse a `CodeChunk` by parsing its `text` into an AST and traversing it."""

    imports: typing.List[str]
    declares: typing.List[typing.Union[Variable, Function]]
    assigns: typing.List[str]
    alters: typing.List[str]
    uses: typing.List[str]
    reads: typing.List[str]

    seen_vars: typing.List[str]

    name_skip: typing.List[str] = []

    def reset(self) -> None:
        """Reset the storage lists."""
        self.imports = []
        self.declares = []
        self.assigns = []
        self.alters = []
        self.uses = []
        self.reads = []

        self.seen_vars = []

    def add_variable(self, name: str, type_annotation: typing.Optional[str]) -> None:
        """
        Store a variable declaration.

        Parses the `type_annotation` (if set) and transforms to a Schema subclass.
        """
        if name in self.seen_vars:
            return

        self.seen_vars.append(name)
        self.declares.append(
            Variable(name, schema=annotation_name_to_schema(type_annotation))
        )

    def add_name(self, name: str, target: typing.List) -> None:
        """
        Add a name to the target list, if it not already used.

        `target` should be one of the self properties (`imports`, `declares`, etc). `seen_vars` is the global record of
        any seen names to prevent duplicates (e.g. a variable would not be both declared [in `declares` list] and then
        used [in `uses` list]).
        """
        if name in self.name_skip:  # we can temporarily skip names we are not interested in (e.g. inside lambdas)
            return

        if name not in self.seen_vars and name not in target:
            self.seen_vars.append(name)
            target.append(name)

    def add_alters(self, name: str) -> None:
        """
        A special function for adding an `alters` name instead of using the general `add_name` function above.

        The difference is that if a name is added here and is already in `uses`, then it is removed from `uses` and
        added to `alters` as this is a more accurate definition.
        """
        if name in self.alters:
            return

        if name in self.seen_vars:
            if name not in self.uses:
                return

            self.uses.remove(name)
            # var has been seen, but only in `uses`, a more accurate description is `alters` so
            # move it there

        self.seen_vars.append(name)
        self.alters.append(name)

    def parse(self, chunk: CodeChunk) -> CodeChunkParseResult:
        """
        Main entry function for this class, parses a CodeChunk's properties into a CodeChunkParseResult.
        """
        self.reset()

        try:
            chunk_ast = ast.parse(chunk.text)
        except SyntaxError as exc:  # This should be either SyntaxError or IndentationError (which is a subclass)
            return CodeChunkParseResult(None, error=exception_to_code_error(exc))

        # If this is True, then there should be a call to 'open' somewhere in the code, which
        # means the parser should try to find it. This is a basic check so there might not be
        # one (like if the code did , but if 'open(' is NOT in the string then there definitely
        # ISN'T one
        search_for_open = 'open(' in chunk.text

        for statement in chunk_ast.body:
            self.parse_statement(statement)

        if search_for_open:
            self._parse_file_reads(chunk_ast)

        return CodeChunkParseResult(chunk_ast, self.imports, self.assigns, self.declares, self.alters, self.uses,
                                    self.reads)

    # pylint: disable=R0912,R0915 # Too many branches/statements warning but this is kind of a special case
    def parse_statement(self,
                        statement: typing.Union[
                            ast.stmt, ast.expr, typing.Sequence[typing.Union[ast.stmt, ast.expr, ast.comprehension]]]
                        ) -> None:
        """General statement parser that delegates to parsers for specific parser types."""
        if isinstance(statement, list):
            for sub_statement in statement:
                self.parse_statement(sub_statement)
        elif isinstance(statement, (ast.ImportFrom, ast.Import)):
            self._parse_import(statement)
        elif isinstance(statement, (ast.Assign, ast.AnnAssign)):
            self._parse_assigns(statement)
        elif isinstance(statement, ast.BinOp):
            self._parse_bin_op(statement)
        elif isinstance(statement, ast.BoolOp):
            self._parse_bool_op(statement)
        elif isinstance(statement, ast.Attribute):
            self._parse_attribute(statement)
        elif isinstance(statement, ast.Subscript):
            self._parse_subscript(statement, False)
        elif isinstance(statement, ast.Call):
            self._parse_call(statement)
        elif isinstance(statement, ast.FunctionDef):
            self._parse_function_def(statement)
        elif isinstance(statement, ast.Dict):
            self._parse_dict(statement)
        elif isinstance(statement, (ast.List, ast.Tuple)):
            self.parse_statement(statement.elts)
        elif isinstance(statement, ast.Name):
            self.add_name(statement.id, self.uses)
        elif isinstance(statement, ast.Expr):
            self.parse_statement(statement.value)
        elif isinstance(statement, ast.AugAssign):
            self._parse_aug_assign(statement)
        elif isinstance(statement, (ast.If, ast.While)):
            self._parse_if_while(statement)
        elif isinstance(statement, ast.Compare):
            self._parse_compare(statement)
        elif isinstance(statement, ast.For):
            self._parse_for(statement)
        elif isinstance(statement, ast.Try):
            self._parse_try(statement)
        elif isinstance(statement, ast.ExceptHandler):
            self._parse_except_handler(statement)
        elif isinstance(statement, ast.With):
            self._parse_with(statement)
        elif isinstance(statement, (ast.ListComp, ast.SetComp)):
            self._parse_list_or_set_comprehension(statement)
        elif isinstance(statement, ast.DictComp):
            self._parse_dict_comprehension(statement)
        elif isinstance(statement, ast.comprehension):
            self._parse_comprehension(statement)
        elif isinstance(statement, ast.Lambda):
            self._parse_lambda(statement)
        elif isinstance(statement, ast.UnaryOp):
            self._parse_unary_op(statement)
        elif isinstance(statement, (
                ast.ClassDef, ast.Num, ast.Str, ast.Pass, ast.NameConstant, ast.Bytes, ast.Break, ast.Continue)):
            pass
        else:
            LOGGER.warning('Unrecognized statement %s', statement)

    def _parse_import(self, statement: typing.Union[ast.ImportFrom, ast.Import]) -> None:
        """
        Parse 'import ...' or 'from ... import ...' statements.

        Adds the modules that are being imported from to the `imports` list.
        """
        if isinstance(statement, ast.ImportFrom):
            modules = [statement.module]
        else:
            modules = [name.name for name in statement.names]

        for module in modules:
            if module is not None and module not in self.imports:
                self.imports.append(module)

    def _parse_reference(self,
                         ref: typing.Optional[typing.Union[ast.stmt, ast.expr, ast.slice]]) -> typing.Optional[str]:
        """
        Attempts to get the final name of a reference (usually an `Index` used in a subscript).

        For example, `a[b:c:d]` -> `b`, `c`, `d`.
        """
        if isinstance(ref, ast.Name):
            return ref.id

        if not isinstance(ref, ast.Subscript) and hasattr(ref, 'value'):
            # for some reason isinstance (ref, ast.Index) doesn't work, so the hasattr check is a
            # workaround to check for an ast.Index object
            return self._parse_reference(ref.value)  # type: ignore

        return None

    def _parse_assigns(self, statement: typing.Union[ast.Assign, ast.AnnAssign]) -> None:
        """
        Parse an assigment and try to choose the best place to store it.

        If the assignment has a type annotation, treat that as a variable declaration (name into `declares`).
        If the  assignment is to an object (e.g. `x.y = z`) or list (e.g. `x[y] = z` treat that as an alters
        (`x` into `alters`).
        Otherwise, put the name into the `assigns` list. Also parse the right side of the expression to find other
        variables that are used.
        """
        if isinstance(statement, ast.Assign):
            targets = statement.targets
        elif isinstance(statement, ast.AnnAssign):
            targets = [statement.target]
        else:
            raise TypeError('{} has no target or targets'.format(statement))

        for target in targets:
            if isinstance(target, ast.Attribute):
                self.add_alters(self._recurse_attribute(target))
                continue

            if isinstance(target, ast.Subscript):
                subscript_name = self._parse_subscript(target, True)
                if subscript_name is not None:
                    self.add_alters(subscript_name)
                continue

            if isinstance(target, ast.Name):
                if isinstance(statement, ast.AnnAssign):
                    annotation_name = statement.annotation.id if isinstance(statement.annotation, ast.Name) else None
                    self.add_variable(target.id, annotation_name)
                else:
                    self.add_name(target.id, self.assigns)

        statement_value = getattr(statement, 'value', None)
        if statement_value is not None:
            self.parse_statement(statement_value)

    def _recurse_attribute(self, ref: typing.Union[ast.stmt, ast.expr]) -> str:
        """Recurse through an attribute to get the actual variable (e.g. `x.y.z` -> `x`)."""
        if hasattr(ref, 'value'):
            if isinstance(ref.value, ast.Attribute):  # type: ignore
                return self._recurse_attribute(ref.value)  # type: ignore

            if isinstance(ref.value, ast.Name):  # type: ignore
                return ref.value.id  # type: ignore

        raise TypeError('Can\'t get name of {}'.format(ref))

    def _parse_attribute(self, statement: ast.Attribute) -> None:
        """
        Parse a standalone attribute and add it to the `uses` list.

        This will be in cases like list or dict literals, or function calls.
        Some examples:
        `a = [b.c]` -> `b` into `uses`.
        `a = {c.d: e.f}` -> `c` and `e` into `uses`.
        `call_func(a.b, c.d)` -> `a` and `c` into `uses`.
        """
        self.add_name(self._recurse_attribute(statement), self.uses)

    def _parse_subscript(self, statement: ast.Subscript, in_assign: bool) -> typing.Optional[str]:
        """
        Parse a subscript (list access).

        Handles single element access (`a[b]`) and slices (`a[b:c:d]`).
        """
        if isinstance(statement.slice, ast.Slice):
            for ref_name in ('lower', 'step', 'upper'):
                ref = getattr(statement.slice, ref_name, None)
                name = self._parse_reference(ref)
                if name is not None:
                    self.add_name(name, self.uses)
                if ref is not None:
                    self.parse_statement(ref)
        else:
            slice_name = self._parse_reference(statement.slice)
            if slice_name is not None:
                self.add_name(slice_name, self.uses)

        value_ref = self._parse_reference(statement.value)
        if value_ref is not None:
            if not in_assign:
                self.add_name(value_ref, self.uses)
            return value_ref

        if isinstance(statement, ast.Subscript) and isinstance(statement.value, ast.Subscript):
            return self._parse_subscript(statement.value, in_assign)

        self.parse_statement(statement.value)
        return None

    def _parse_bin_op(self, statement: ast.BinOp) -> None:
        """Parse a binary operation, e.g `a + b`, `c - d`."""
        self.parse_statement(statement.left)
        self.parse_statement(statement.right)

    def _parse_bool_op(self, statement: ast.BoolOp) -> None:
        """Pares a boolean operation, e.g. `a or b`, `d or e`."""
        self.parse_statement(statement.values)

    def _parse_call(self, statement: ast.Call) -> None:
        """Parse a function call to extract the variables used."""
        if hasattr(statement, 'args'):
            self.parse_statement(statement.args)

        if hasattr(statement, 'keywords'):
            for keyword in statement.keywords:
                self.parse_statement(keyword.value)

    def _parse_function_def(self, statement: ast.FunctionDef) -> None:
        """Parse a function definition to extract the `Parameter`s it accepts."""
        if statement.name in self.seen_vars:
            return

        return_ann = statement.returns.id if isinstance(statement.returns, ast.Name) else None

        func = Function(statement.name, returns=annotation_name_to_schema(return_ann), parameters=[])

        for i, arg in enumerate(statement.args.args):
            param = Parameter(arg.arg)

            if arg.annotation and hasattr(arg.annotation, 'id'):
                param.schema = annotation_name_to_schema(arg.annotation.id)  # type: ignore

            default_index = len(statement.args.defaults) - len(statement.args.args) + i
            # Only the last len(statement.args.defaults) can have defaults (since they must come after non-default
            # parameters)
            if default_index >= 0:
                default = statement.args.defaults[default_index]

                if isinstance(default, ast.Num):
                    param.default = default.n
                elif isinstance(default, ast.Str):
                    param.default = default.s
                elif isinstance(default, ast.NameConstant):
                    # default of None/True/False
                    param.default = default.value
                else:
                    self.parse_statement(default)
                param.required = False
            else:
                param.required = True

            func.parameters.append(param)

        if statement.args.vararg:
            func.parameters.append(Parameter(statement.args.vararg.arg, required=False, repeats=True))

        if statement.args.kwarg:
            func.parameters.append(Parameter(statement.args.kwarg.arg, required=False, extends=True))

        self.seen_vars.append(func.name)
        self.declares.append(func)

    def _parse_dict(self, statement: ast.Dict) -> None:
        """
        Parse a dictionary definition, adding any variables it uses as keys or values
        to `uses`.
        """
        for key in statement.keys:
            if isinstance(key, ast.Name):
                self.add_name(key.id, self.uses)
            elif key:
                self.parse_statement(key)
        for value in statement.values:
            if isinstance(value, ast.Name):
                self.add_name(value.id, self.uses)
            else:
                self.parse_statement(value)

    def _parse_aug_assign(self, statement: ast.AugAssign) -> None:
        """
        Parse an augmented assignment (e.g. a += 1).

        Adds the variable being altered to the `alters` list.
        """
        if isinstance(statement.target, ast.Name):
            self.add_alters(statement.target.id)
        elif isinstance(statement.target, ast.Attribute):
            self.add_alters(self._recurse_attribute(statement.target))
        elif isinstance(statement.target, ast.Subscript):
            target_name = self._parse_subscript(statement.target, True)
            if target_name:
                self.add_alters(target_name)

        self.parse_statement(statement.value)

    def _parse_if_while(self, statement: typing.Union[ast.If, ast.While]) -> None:
        """Parse the test (condition), body, and `elif`/`else` statements of an `if` or `while`."""
        self.parse_statement(statement.test)
        self.parse_statement(statement.body)
        self.parse_statement(statement.orelse)

    def _parse_compare(self, statement: ast.Compare) -> None:
        """Parse a comparison statement (e.g. a > b, c < d, etc) and add the variables it uses to the `uses` list."""
        self.parse_statement(statement.left)
        self.parse_statement(statement.comparators)

    def _parse_for(self, statement: ast.For) -> None:
        """
        Parse a `for ...:` statement.

        Since the variable being assigned in iteration is available after the loop, it is added to the `assigns` list.
        """
        if isinstance(statement.target, ast.Name):
            self.add_name(statement.target.id, self.assigns)
        self.parse_statement(statement.iter)
        self.parse_statement(statement.body)
        self.parse_statement(statement.orelse)

    def _parse_try(self, statement: ast.Try) -> None:
        """Parse a `try`/`except`/`finally`/`else` statement."""
        self.parse_statement(statement.handlers)  # type: ignore  # Doesn't seem to understand List[ExceptHandler]
        # is valid
        self.parse_statement(statement.body)
        self.parse_statement(statement.finalbody)
        self.parse_statement(statement.orelse)

    def _parse_except_handler(self, statement: ast.ExceptHandler) -> None:
        """Parse an `except` handler (i.e. parse the statements in its `body`)."""
        self.parse_statement(statement.body)

    def _parse_with(self, statement: ast.With) -> None:
        """Parse a `with` statement (i.e. parse the statements in its `body`)."""
        self.parse_statement(statement.body)

    def _parse_list_or_set_comprehension(self, statement: typing.Union[ast.ListComp, ast.SetComp]) -> None:
        """Parse a list or set comprehension (they have the same interface)."""
        if not isinstance(statement.elt, ast.Name):
            # skip simple name assigns since they aren't really usable after the loop
            self.parse_statement(statement.elt)
        self.parse_statement(statement.generators)

    def _parse_dict_comprehension(self, statement: ast.DictComp) -> None:
        """Parse a dict comprehension."""
        self.parse_statement(statement.generators)
        self.parse_statement(statement.value)

    def _parse_comprehension(self, statement: ast.comprehension) -> None:
        """Parse a generator such as used in a comprehension."""
        target = statement.target

        if hasattr(target, 'elts'):
            self.name_skip = [name.id for name in target.elts]  # type: ignore
        elif hasattr(target, 'id'):
            self.name_skip = [target.id]  # type: ignore

        self.parse_statement(statement.iter)
        self.parse_statement(statement.ifs)

        self.name_skip = []

    def _parse_lambda(self, statement: ast.Lambda) -> None:
        # add the arguments to the name_skip so that they won't be added to 'uses' in the body parse
        self.name_skip = [arg.arg for arg in statement.args.args]  # I feel like a pirate

        self.parse_statement(statement.body)

        self.name_skip = []

    def _parse_unary_op(self, statement: ast.UnaryOp) -> None:
        self.parse_statement(statement.operand)

    def _parse_file_reads(self, chunk_ast: ast.Module) -> None:
        """
        Walk the ast (including into function defs) and look for `open` function calls.

        Add any file reads (any `open` calls that aren't exclusively writes) to the `reads` list.
        """
        for node in ast.walk(chunk_ast):
            if isinstance(node, ast.Call) and isinstance(getattr(node, 'func', None),
                                                         ast.Name) and node.func.id == 'open':  # type: ignore
                filename = parse_open_filename(node)

                if filename and filename not in self.reads:
                    self.reads.append(filename)



"""
__main__.py can be executed (once this library is installed) like this:

python3 -m stencila.pyla execute <inputfile> <outputfile> [parameters]

See README.md for more information.

Warning: `eval` and `exec` are used to run code in the document. Don't execute documents that you haven't verified
yourself.
"""

import logging
from sys import argv, stderr

from .system import register, deregister
from .interpreter import Interpreter, execute_from_cli
from .servers import StdioServer


def cli_execute():
    """Execute an executable document, delegating to the execute_from_cli function."""
    execute_from_cli(argv[2:])


def cli_compile():
    """Compile an executable document by delegating to the execute_from_cli function with the `compile_only` flag."""
    execute_from_cli(argv[2:], True)

def main():
    """The main entry point to this module, read the first CLI arg and call out to the corresponding function."""
    command = argv[1] if len(argv) > 1 else ''

    # Send logs to stderr so that there it does not interfere with
    # JSON-RPC comms using length-prefixed streams over stdio.
    logging.basicConfig(stream=stderr, level=logging.DEBUG)

    if command == 'execute':
        cli_execute()
    elif command == 'compile':
        cli_compile()
    elif command == 'serve':
        StdioServer(Interpreter()).start()
    elif command == 'register':
        register()
    elif command == 'deregister':
        deregister()
    else:
        stderr.write('Unknown command "{}"\n'.format(command))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from setuptools import setup

HERE = os.path.dirname(__file__)

sys.path.insert(0, os.path.join(HERE, 'stencila'))
from pyla import __version__

with io.open(os.path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='stencila-pyla',
    description='Python interpreter for Stencila',
    version=__version__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Stencila and contributors',
    author_email='hello@stenci.la',
    python_requires='>=3.6.0',
    url='https://github.com/stencila/pyla',
    packages=['stencila.pyla'],
    install_requires=[
        'astor==0.8.0',
        'stencila-schema==0.29.0'
    ],
    extras_require={},
    include_package_data=True,
    license='Apache-2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ]
)



"""
This is for interpreting and executing an executable document in JSON format. It can be called from the command line:

python3 interpreter.py <infile> <outfile> [document parameters]

See README.md for more information.

Warning: `eval` and `exec` are used to run code in the document. Don't execute documents that you haven't verified
yourself.
"""

import sys

import argparse
import base64
import datetime
import json
import logging
import typing
from contextlib import redirect_stdout
from io import TextIOWrapper, BytesIO

import ast
import astor
from stencila.schema.types import Node, Parameter, CodeChunk, Article, Entity, CodeExpression, \
    ConstantSchema, EnumSchema, BooleanSchema, NumberSchema, IntegerSchema, StringSchema, \
    ArraySchema, TupleSchema, ImageObject, Datatable, DatatableColumn, Function
from stencila.schema.util import from_json, to_json

from .errors import CapabilityError
from .code_parsing import CodeChunkExecution, set_code_error, CodeChunkParser, simple_code_chunk_parse, \
    CodeChunkParseResult

try:
    import matplotlib.figure
    import matplotlib.artist
    from matplotlib.cbook import silent_list

    # pylint: disable=C0103
    MPLFigure = matplotlib.figure.Figure
    # pylint: disable=C0103
    MPLArtist = matplotlib.artist.Artist
    MPL_AVAILABLE = True
except ImportError:
    # pylint: disable=R0903
    class MPLFigure:  # type: ignore
        """A fake MPLFigure to prevent ReferenceErrors later."""


    # pylint: disable=R0903
    class MLPArtist:  # type: ignore
        """A fake MLPArtist to prevent ReferenceErrors later."""


    # pylint: disable=R0903,C0103
    class silent_list:  # type: ignore
        """A fake silent_list to prevent ReferenceErrors later."""


    MPL_AVAILABLE = False

try:
    from pandas import DataFrame
    import numpy

    PANDAS_AVAILABLE = True
except ImportError:
    # pylint: disable=R0903
    class DataFrame:  # type: ignore
        """A fake DataFrame to prevent ReferenceErrors later."""


    PANDAS_AVAILABLE = False

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

CHUNK_PREVIEW_LENGTH = 20

ExecutableCode = typing.Union[CodeChunkExecution, CodeExpression]
StatementRuntime = typing.Tuple[bool, typing.Union[str], typing.Callable]

# Used to indicate that a particular output should not be added to outputs (c.f. a valid `None` value)
SKIP_OUTPUT_SEMAPHORE = object()


class CodeTimer:
    """Context handler for timing code, use inside a `with` statement."""
    _start_time: datetime.datetime
    duration: typing.Optional[datetime.timedelta] = None

    def __enter__(self):
        self.duration = None
        self._start_time = datetime.datetime.now()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.duration = datetime.datetime.now() - self._start_time

    @property
    def duration_seconds(self) -> float:
        """Calculate the duration (time between `__enter__` and `__exit__` in microseconds."""
        if not self.duration:
            raise RuntimeError('CodeTimer has not yet been run')

        return self.duration.total_seconds()


class StdoutBuffer(TextIOWrapper):
    """Used for capturing output to stdout."""

    def write(self, string: typing.Union[bytes, str]) -> int:
        if isinstance(string, str):
            return super(StdoutBuffer, self).write(string)
        return super(StdoutBuffer, self).buffer.write(string)


class DocumentCompilationResult(typing.NamedTuple):
    """Stores references to Parameters and Code that is parsed from a Document."""

    parameters: typing.List[Parameter] = []
    code: typing.List[ExecutableCode] = []


class DocumentCompiler:
    """Parse an executable document (`Article`) and cache references to its parameters and code nodes."""

    TARGET_LANGUAGE = 'python'

    function_depth: int = 0

    def compile(self, source: Article) -> DocumentCompilationResult:
        """
        Compile an `Article` by walking it and finding `Parameter`s and `CodeChunk` or `CodeExpression` nodes.

        These are set on a `DocumentCompilationResult` which can be passed to the `Interpreter`.
        """
        self.function_depth = 0
        dcr = DocumentCompilationResult()

        self.handle_item(source, dcr)
        return dcr

    def handle_item(self, item: typing.Any, compilation_result: DocumentCompilationResult) -> None:
        """
        Parse any kind of dict, list of Entity.

        Returns nothing but updates the passed in `DocumentCompilationResult`.
        """
        if isinstance(item, dict):
            self.traverse_dict(item, compilation_result)
        elif isinstance(item, list):
            self.traverse_list(item, compilation_result)
        elif isinstance(item, Entity):
            if isinstance(item, (CodeChunk, CodeExpression)):
                if item.programmingLanguage == self.TARGET_LANGUAGE:  # Only add Python code
                    self.handle_code(item, compilation_result)

            elif isinstance(item, Parameter) and self.function_depth == 0:
                compilation_result.parameters.append(item)
                LOGGER.debug('Adding %s', type(item))

            if isinstance(item, Function):
                # This prevents treating a `Parameter` found inside `Function` as a document parameter
                self.function_depth += 1

            self.traverse_dict(item.__dict__, compilation_result)

            if isinstance(item, Function):
                self.function_depth -= 1

    @staticmethod
    def handle_code(item: typing.Union[CodeChunk, CodeExpression],
                    compilation_result: DocumentCompilationResult) -> None:
        """Parse a CodeChunk or CodeExpression and add it to `compilation_result.code` list."""
        if isinstance(item, CodeChunk):
            cc_result, item = Interpreter.compile_code_chunk(item)

            code_to_add = CodeChunkExecution(item, cc_result)
        else:
            try:
                ast.parse(item.text)
            except SyntaxError as exc:  # Probably will only be this
                set_code_error(item, exc)
            code_to_add = item
        compilation_result.code.append(code_to_add)
        LOGGER.debug('Adding %s', type(item))

    def traverse_dict(self, _dict: dict, compilation_result: DocumentCompilationResult) -> None:
        """Traverse into each item of a dict."""
        for child in _dict.values():
            self.handle_item(child, compilation_result)

    def traverse_list(self, _list: typing.List, compilation_result: DocumentCompilationResult) -> None:
        """Traverse into each element in a list."""
        for child in _list:
            self.handle_item(child, compilation_result)


class Interpreter:
    """Execute a list of code blocks, maintaining its own `globals` scope for this execution run."""

    """
    List of values for the `programmingLanguage` property that are handled
    by this interpreter.
    """
    PROGRAMMING_LANGUAGES = ['py', 'python']

    """
    JSON Schema specification of the types of nodes that this intertpreter's
    `compile` and `execute` methods are capable of handling
    """
    CODE_CAPABILITIES = {
        'type': 'object',
        'required': ['node'],
        'properties': {
            'node': {
                'type': 'object',
                'required': ['type', 'programmingLanguage'],
                'properties': {
                    'type': {
                        'enum': ['CodeChunk', 'CodeExpression']
                    },
                    'programmingLanguage': {
                        'enum': PROGRAMMING_LANGUAGES
                    }
                }
            }
        }
    }

    """
    The manifest of this interpreter's capabilities and addresses.

    Conforms to Executa's
    [Manifest](https://github.com/stencila/executa/blob/v1.4.0/src/base/Executor.ts#L63)
    interface.
    """
    MANIFEST = {
        'version': 1,
        'capabilities': {
            'compile': CODE_CAPABILITIES,
            'execute': CODE_CAPABILITIES
        },
        'addresses': {
            'stdio': {
                'type': 'stdio',
                'command': sys.executable,
                'args': ['-m', 'stencila.pyla', 'serve']
            }
        }
    }

    globals: typing.Dict[str, typing.Any]
    locals: typing.Dict[str, typing.Any]

    def __init__(self) -> None:
        self.globals = {}
        self.locals = {}

    @staticmethod
    def compile_code_chunk(chunk: CodeChunk) -> typing.Tuple[CodeChunkParseResult, CodeChunk]:
        """
        Compile a `CodeChunk`.

        Returns a `CodeChunkParseResult` which is primarily needed for the AST, and the `CodeChunk` itself, which has
        its code metadata properties set.
        """
        parser = CodeChunkParser()
        cc_result = parser.parse(chunk)
        chunk.imports = cc_result.combined_code_imports(chunk.imports)
        chunk.declares = cc_result.declares
        chunk.assigns = cc_result.assigns
        chunk.alters = cc_result.alters
        chunk.uses = cc_result.uses
        chunk.reads = cc_result.reads

        if cc_result.error:
            set_code_error(chunk, cc_result.error)
        return cc_result, chunk

    @staticmethod
    def compile(node: Node) -> Node:
        """Compile a `CodeChunk`"""
        if isinstance(node, CodeChunk) and Interpreter.is_python_code(node):
            return Interpreter.compile_code_chunk(node)[1]
        raise CapabilityError('compile', node=node)

    def execute(self, node: Node, parameter_values: typing.Dict[str, typing.Any] = None) -> Node:
        """Execute a `CodeChunk` or `CodeExpression`"""
        _locals = self.locals
        if parameter_values is not None:
            _locals.update(parameter_values)

        if isinstance(node, CodeExpression):
            return self.execute_code_expression(node, _locals)
        if isinstance(node, CodeChunk):
            cce = simple_code_chunk_parse(node)
            return self.execute_code_chunk(cce, _locals)
        if isinstance(node, CodeChunkExecution):
            return self.execute_code_chunk(node, _locals)
        raise CapabilityError('execute', node=node)

    @staticmethod
    def is_python_code(code: typing.Union[CodeChunk, CodeExpression]) -> bool:
        """Is a `CodeChunk` or `CodeExpression` Python code?"""
        return code.programmingLanguage.lower() in Interpreter.PROGRAMMING_LANGUAGES

    def execute_code_expression(self, expression: CodeExpression,
                                _locals: typing.Dict[str, typing.Any]) -> CodeExpression:
        """Evaluate `CodeExpression.text`, and get the result. Catch any exception the occurs."""
        try:
            # pylint: disable=W0123  # Disable warning that eval is being used.
            expression.output = self.decode_output(eval(expression.text, self.globals, _locals))
        # pylint: disable=W0703  # we really don't know what Exception some eval'd code might raise.
        except Exception as exc:
            set_code_error(expression, exc)

        return expression

    def execute_code_chunk(self, chunk_execution: CodeChunkExecution,
                           _locals: typing.Dict[str, typing.Any]) -> CodeChunk:
        """Execute a `CodeChunk` that has been parsed and stored in a `CodeChunkExecution`."""
        chunk, parse_result = chunk_execution

        if parse_result.chunk_ast is None:
            LOGGER.info('Not executing CodeChunk without AST: %s', chunk.text[:CHUNK_PREVIEW_LENGTH])
            return chunk

        cc_outputs: typing.List[typing.Any] = []

        duration = 0.0

        for statement in parse_result.chunk_ast.body:
            duration, error_occurred = self.execute_statement(statement, chunk, _locals, cc_outputs, duration)

            if error_occurred:
                break  # stop executing the rest of the statements in the chunk after capturing the outputs

        chunk.duration = duration

        if MPL_AVAILABLE:
            # Because of the way matplotlib might progressively build an image, only keep the last MPL that was
            # generated for a specific code chunk
            mpl_output_indexes = [i for i, output in enumerate(cc_outputs) if self.value_is_mpl(output)]

            if mpl_output_indexes:
                for i in reversed(mpl_output_indexes[:-1]):
                    # remove all but the last mpl
                    cc_outputs.pop(i)

                new_last_index = mpl_output_indexes[-1] - (len(mpl_output_indexes) - 1)  # output will have shifted back

                cc_outputs[new_last_index] = self.decode_mpl()

        chunk.outputs = cc_outputs

        return chunk

    def execute_statement(self, statement: ast.stmt, chunk: CodeChunk, _locals: typing.Dict[str, typing.Any],
                          cc_outputs: typing.List[str], duration: float) -> typing.Tuple[float, bool]:
        """
        Execute a single AST statement.

        The statement will be executed with `eval` or `exec` depending on its type (see `parse_statement_runtime`).
        """
        error_occurred = False

        capture_result, code_to_run, run_function = self.parse_statement_runtime(statement)
        stdout = StdoutBuffer(BytesIO(), sys.stdout.encoding)
        result = None

        with redirect_stdout(stdout):
            try:
                with CodeTimer() as code_timer:
                    result = run_function(code_to_run, self.globals, _locals)
                duration += code_timer.duration_seconds
            # pylint: disable=W0703  # we really don't know what Exception some exec'd code might raise.
            except Exception as exc:
                error_occurred = True
                set_code_error(chunk, exc)

        if capture_result and result is not None:
            self.add_output(cc_outputs, result)

        stdout.seek(0)
        std_out_output = stdout.buffer.read()
        if std_out_output:
            cc_outputs.append(std_out_output.decode('utf8'))

        # could save a variable by returning `duration` as `None` if an error occurs but I think that breaks readability
        return duration, error_occurred

    def add_output(self, cc_outputs: typing.List[typing.Any], result: typing.Any) -> None:
        """
        Add an output to cc_outputs.

        Should be inside `execute_statement` as it's only used there, but pylint complains about too many local
        variables.
        """
        decoded = self.decode_output(result)
        if decoded != SKIP_OUTPUT_SEMAPHORE:
            cc_outputs.append(decoded)

    @staticmethod
    def parse_statement_runtime(statement: ast.stmt) -> StatementRuntime:
        """
        Determine the information to execute a statement.

        Returns the function with which to execute a statement (`eval` or `exec`), whether its output should be captured
        or not, and the compiled code to execute.

        See the other comments below on how/why these functions are chosen.
        """

        if isinstance(statement, ast.Expr):
            # An expression is something that we want to capture the result of - this could be something like
            # `a + 3` or a function call (not an assignment). It must be executed with `eval`. Since `eval` can't
            # execute compiled code, `astor` is used to convert it back to source code.
            capture_result = True
            run_function = eval
            code_to_run = astor.to_source(statement)
        else:
            # We don't care about the result of this call (it could just be an assignment, update or even function
            # definition) so it can be executed with `exec`.
            capture_result = False
            run_function = exec
            mod = ast.Module()
            mod.body = [statement]
            code_to_run = compile(mod, '<ast>', 'exec')
        return capture_result, code_to_run, run_function

    @staticmethod
    def value_is_mpl(value: typing.Any) -> bool:
        """Basic type checking to determine if a variable is a MatPlotLib figure."""
        if not MPL_AVAILABLE:
            return False

        return isinstance(value, (MPLFigure, MPLArtist)) or (isinstance(value, list) and
                                                             len(value) == 1 and
                                                             isinstance(value[0], MPLArtist))

    @staticmethod
    def decode_mpl() -> ImageObject:
        """
        Decode a matplotlib `MPLFigure` or `MPLArtist` into an `ImageObject`.

        The `matplotlib.pyplot.savefig` function just saves the current MPL figure that's in the context.
        """
        image = BytesIO()
        matplotlib.pyplot.savefig(image, format='png')
        src = 'data:image/png;base64,' + base64.encodebytes(image.getvalue()).decode()
        return ImageObject(src)

    @staticmethod
    def decode_dataframe(data_frame: DataFrame) -> Datatable:
        """Decode a pandas `DataFrame` into a `Datatable`"""
        columns = []

        for column_name in data_frame.columns:
            column = data_frame[column_name]
            values = column.tolist()
            if column.dtype in (numpy.bool_, numpy.bool8):
                schema = BooleanSchema()
                values = [bool(row) for row in values]
            elif column.dtype in (numpy.int8, numpy.int16, numpy.int32, numpy.int64):
                schema = IntegerSchema()
                values = [int(row) for row in values]
            elif column.dtype in (numpy.float16, numpy.float32, numpy.float64):
                schema = NumberSchema()
                values = [float(row) for row in values]
            elif column.dtype in (numpy.str_, numpy.unicode_,):
                schema = StringSchema()
            else:
                schema = None

            columns.append(
                DatatableColumn(column_name, values, schema=ArraySchema(items=schema))
            )

        return Datatable(columns)

    def decode_output(self, output: typing.Any) -> typing.Any:
        """
        Check if the output is convertible from a special data type to a Stencila type.

        If not, just return the original object."""

        if MPL_AVAILABLE:
            if isinstance(output, silent_list):
                return SKIP_OUTPUT_SEMAPHORE

        if isinstance(output, list):
            return [self.decode_output(item) for item in output]

        if isinstance(output, tuple):
            return tuple(self.decode_output(item) for item in output)

        if isinstance(output, DataFrame):
            return self.decode_dataframe(output)

        if PANDAS_AVAILABLE:
            if isinstance(output, numpy.ndarray):
                return output.tolist()

        return output


class ParameterParser:
    """
    Parse parameters that the document requires, from the command line.

    The `ArgumentParser` will fail if any required parameters are not passed in.
    """
    parameters: typing.Dict[str, Parameter]
    parameter_values: typing.Dict[str, typing.Any]

    def __init__(self, parameters: typing.List[Parameter]) -> None:
        self.parameters = {parameter.name: parameter for parameter in parameters}
        self.parameter_values = {}

    def parse_cli_args(self, cli_args: typing.List[str]) -> None:
        """
        Parse a list of `cli_args` (strings) into the `parameter_values` dict.

        This will fill in defaults and convert types based on the Parameter's schema. If a required argument is missing
        then the `ArgumentParser` class will cause the process to exit.
        """
        if not self.parameters:
            LOGGER.debug('No parameters passed to parse_cli_args')
            return

        param_parser = argparse.ArgumentParser(description='Parse Parameters')

        for param in self.parameters.values():
            if not isinstance(param.schema, ConstantSchema):
                param_parser.add_argument('--' + param.name, dest=param.name, required=param.required)

        args, _ = param_parser.parse_known_args(cli_args)

        LOGGER.debug('Parsed command line args: %s', args)

        for param_name in self.parameters:
            cli_value = getattr(args, param_name, None)
            if not cli_value and self.parameters[param_name].default:
                self.parameter_values[param_name] = self.parameters[param_name].default
            else:
                self.parameter_values[param_name] = self.deserialize_parameter(self.parameters[param_name], cli_value)

    @staticmethod
    def deserialize_parameter(parameter: Parameter, value: typing.Any) -> typing.Any:
        """Convert a value (usually a string) into the type specified by the parameter's schema."""
        if isinstance(parameter.schema, ConstantSchema):
            # A ConstantSchema doesn't take a value it stores its own value.
            return parameter.schema.value

        if isinstance(parameter.schema, EnumSchema):
            if value not in parameter.schema.values:
                raise TypeError('{} not found in enum values for {}'.format(value, parameter.name))
            return value

        if isinstance(parameter.schema, BooleanSchema):
            return value.lower() in ('true', 'yes', '1', 't')

        conversion_function: typing.Optional[typing.Callable] = None

        if isinstance(parameter.schema, IntegerSchema):
            conversion_function = int
        elif isinstance(parameter.schema, NumberSchema):
            conversion_function = float  # If there are problems with inaccuracy consider using Decimal instead
        elif isinstance(parameter.schema, (ArraySchema, TupleSchema)):
            conversion_function = json.loads

        return conversion_function(value) if conversion_function else value


def read_input(input_file: str) -> Article:
    """Read an input file from a "path", decode it as JSON and return the result.

    If the path is "-" then read from stdin. If the decoded object is not an `Article` then raise a `TypeError`.
    """
    if input_file == '-':
        j = sys.stdin.read()
    else:
        with open(input_file) as i:
            j = i.read()
    article = from_json(j)
    if not isinstance(article, Article):
        raise TypeError('Decoded JSON was not an Article')
    return article


def write_output(output_file: str, article: Article) -> None:
    """
    Convert an article to JSON and write to a "path".

    If the path is "-" then output to stdout.
    """
    if output_file == '-':
        sys.stdout.write(to_json(article))
    else:
        with open(output_file, 'w') as file:
            file.write(to_json(article))


def execute_compilation(compilation_result: DocumentCompilationResult, parameter_flags: typing.List[str]) -> None:
    """Compile an `Article`, and interpret it with the given parameters (in a format that would be read from CLI)."""
    param_parser = ParameterParser(compilation_result.parameters)
    param_parser.parse_cli_args(parameter_flags)
    interpreter = Interpreter()
    for code in compilation_result.code:
        interpreter.execute(code, param_parser.parameter_values)


def compile_article(article: Article) -> DocumentCompilationResult:
    """Compile an `Article`, without executing it afterward."""
    return DocumentCompiler().compile(article)


def execute_from_cli(cli_args: typing.List[str], compile_only=False) -> None:
    """Read arguments from the CLI then parse an `Article` in JSON format, execute it, and output it somewhere."""
    cli_parser = argparse.ArgumentParser()
    cli_parser.add_argument('input_file', help='File to read from or "-" to read from stdin', nargs='?', default='-')
    cli_parser.add_argument('output_file', help='File to write to or "-" to write to stdout', nargs='?', default='-')

    args, _ = cli_parser.parse_known_args(cli_args)

    input_file = args.input_file
    output_file = args.output_file

    article = read_input(input_file)

    compilation_result = compile_article(article)

    if not compile_only:
        execute_compilation(compilation_result, cli_args)

    write_output(output_file, article)



"""
Module for server classes.
"""

import sys

import enum
import json
import logging
import typing
from socket import socket
from stencila.schema.types import Node
from stencila.schema.util import from_dict, object_encode

from .errors import CapabilityError
from .interpreter import Interpreter

StreamType = typing.Union[typing.BinaryIO, socket]


def rpc_json_object_encode(node: Node) -> typing.Union[dict, str]:
    """Like `stencila.schema.util.object_encode` but with support for JsonRpcError."""
    if isinstance(node, JsonRpcError):
        return {
            'code': node.code.value,
            'message': str(node),
            'data': node.data
        }

    return object_encode(node)


def to_json(node: Node) -> str:
    """Convert a node including JsonRrpcErrors, to JSON"""
    return json.dumps(node, default=rpc_json_object_encode, indent=2)


def data_to_bytes(data: typing.Any) -> bytes:
    """Convert `data` to `bytes`."""
    return bytes((data,))


def encode_int(number: int) -> bytes:
    """Pack `number` into varint bytes"""
    buf = b''
    while True:
        to_write = number & 0x7f
        number >>= 7
        if number:
            buf += data_to_bytes(to_write | 0x80)
        else:
            buf += data_to_bytes(to_write)
            break
    return buf


def read_one(stream: StreamType) -> int:
    """Read a byte from the file (as an integer).

    Raises EOFError if the stream ends while reading bytes.
    """
    char = stream_read(stream, 1)
    if not char:
        raise EOFError('Unexpected EOF while reading bytes')
    return ord(char)


def read_length_prefix(stream: StreamType) -> int:
    """Read a varint from `stream`"""
    shift = 0
    result = 0
    while True:
        i = read_one(stream)
        result |= (i & 0x7f) << shift
        shift += 7
        if not i & 0x80:
            break

    return result


def get_stream_buffer(stream: typing.BinaryIO) -> typing.BinaryIO:
    """Get the buffer from a stream, if it exists."""
    buffer = getattr(stream, 'buffer', None)
    return buffer if buffer else stream


def io_read(stream: typing.BinaryIO, count: int) -> bytes:
    """Read `count` bytes from `stream` or its underlying buffer if it exists."""
    return get_stream_buffer(stream).read(count)


def io_write(stream: typing.BinaryIO, message: bytes) -> None:
    """Write to `stream` or its underlying buffer if it exists."""
    stream = get_stream_buffer(stream)
    stream.write(message)
    stream.flush()


def stream_read(stream: StreamType, count: int) -> bytes:
    """Abstract reading from stream to work with IO (buffered/unbuffered) and sockets."""
    if isinstance(stream, socket):
        return stream.recv(count)

    return io_read(stream, count)


def stream_write(stream: StreamType, message: bytes) -> None:
    """Abstract writing to stream to work with IO (buffered/unbuffered) and sockets."""
    if isinstance(stream, socket):
        stream.send(message)
    else:
        io_write(stream, message)


def message_read(stream: StreamType) -> str:
    """Read a length-prefixed message from the stream"""
    message_len = read_length_prefix(stream)
    return stream_read(stream, message_len).decode('utf8')


def message_write(stream: StreamType, message: str) -> None:
    """Write a length-prefixed message to the stream."""
    bites = message.encode('utf8')
    stream_write(stream, encode_int(len(bites)))
    stream_write(stream, bites)


class JsonRpcErrorCode(enum.Enum):
    """
    Error codes defined in JSON-RPC 2.0

    Codes -32000 to -32099	are reserved for implementation-defined server-errors.

    Python implementation of Executa's
    [JsonRpcErrorCode](https://github.com/stencila/executa/blob/v1.0.0/src/base/JsonRpcError.ts#L14).
    """

    """
    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.
    """
    ParseError = -32700

    """The JSON sent is not a valid Request object."""
    InvalidRequest = -32600

    """The method does not exist / is not available."""
    MethodNotFound = -32601

    """Invalid method parameter(s)."""
    InvalidParams = -32602

    """Internal JSON-RPC error."""
    InternalError = -32603

    """Generic server error."""
    ServerError = -32000

    """Capability error."""
    CapabilityError = -32005


class JsonRpcError(Exception):
    """
    A JSON-RPC error that may be part of a response

    Python implementation of Executa's
    [JsonRpcError](https://github.com/stencila/executa/blob/v1.0.0/src/base/JsonRpcError.ts).
    """

    """
    A number that indicates the error type that occurred.
    This MUST be an integer.
    """
    code: JsonRpcErrorCode

    """
    A primitive or structured value that contains additional information about the error.
    This may be omitted.
    The value of this member is defined by the server (e.g. detailed error information,
    nested errors etc.).
    """
    data: typing.Any

    def __init__(self, code: JsonRpcErrorCode, message: str, data: typing.Any = None):
        super().__init__(message)
        self.code = code
        self.data = data


class StreamServer:
    """
    A server that communicates using length-prefixed JSON-RPC messages over streams or sockets.

    Python implementation of Executa's
    [StreamServer](https://github.com/stencila/executa/blob/v1.0.0/src/stdio/StreamServer.ts#L10)
    """

    interpreter: Interpreter
    input_stream: StreamType
    output_stream: StreamType

    def __init__(self, interpreter: Interpreter, input_stream: StreamType, output_stream: StreamType) -> None:
        self.interpreter = interpreter
        self.input_stream = input_stream
        self.output_stream = output_stream

    def read_message(self) -> typing.Iterable[str]:
        """Read a length-prefixed message from the input stream then repeat."""
        while True:
            yield message_read(self.input_stream)

    def write_message(self, message: str) -> None:
        """Write a length-prefixed message to the output stream."""
        message_write(self.output_stream, message)

    def receive_message(self, message: str) -> str:
        """
        Receive a JSON-RPC request and send back a JSON-RPC response.

        The response may have a JSON-RPC `error` if the request was bad.

        Python implementation of Executa's
        [Server.receive](https://github.com/stencila/executa/blob/v1.0.1/src/base/Server.ts#L61).
        """
        request_id = None
        result = None
        error = None

        try:
            try:
                request = json.loads(message)
            except Exception as exc:
                raise JsonRpcError(JsonRpcErrorCode.ParseError, 'Parse error: {}'.format(exc))

            request_id = request.get('id')
            method = request.get('method')
            params = request.get('params')

            if method == 'manifest':
                result = Interpreter.MANIFEST
            elif method in ('compile', 'execute'):
                node = params.get('node')
                if node is None:
                    raise JsonRpcError(JsonRpcErrorCode.InvalidParams, 'Invalid params: "node" is missing')
                node = from_dict(node)
                result = self.interpreter.compile(node) if method == 'compile' else self.interpreter.execute(node)
            else:
                raise JsonRpcError(JsonRpcErrorCode.MethodNotFound, 'Method not found: {}'.format(method))
        except JsonRpcError as exc:
            error = exc
        except CapabilityError as exc:
            error = JsonRpcError(JsonRpcErrorCode.CapabilityError, 'Capability error: {}'.format(exc))
        except Exception as exc:  # pylint: disable=broad-except
            logging.exception(exc)
            error = JsonRpcError(JsonRpcErrorCode.ServerError, 'Internal error: {}'.format(exc))

        response: typing.Dict[str, typing.Any] = {
            'jsonrpc': '2.0',
            'id': request_id,
            'result': result,
            'error': error
        }

        return to_json(response)

    def start(self) -> None:
        """
        Run the server in a loop forever (since the `read_message` generator never finishes).
        """
        for message in self.read_message():
            response = self.receive_message(message)
            self.write_message(response)


class StdioServer(StreamServer):
    """
    A `StreamServer` that uses `stdio` as the transport.

    Python implementation of Executa's
    [StdioServer](https://github.com/stencila/executa/blob/v1.0.0/src/stdio/StdioServer.ts#L12)
    """

    def __init__(self, interpreter: Interpreter):
        super().__init__(interpreter, sys.stdin.buffer, sys.stdout.buffer)



"""Custom error classes"""

class CapabilityError(Exception):
    """
    Custom error class to indicate that an executor is not capable of performing a method call.

    Python implementation of Executa's
    [CapabilityError](https://github.com/stencila/executa/blob/v1.4.0/src/base/errors.ts#L57).
    Is translated to a JSON-RPC error with code `CapabilityError`.
    """

    def __init__(self, method: str, **kwargs):
        params = ', '.join(['{} = {}'.format(name, value) for name, value in kwargs.items()])
        super().__init__('Incapable of method "{}" with params "{}"'.format(method, params))
